#!/usr/bin/env python

import asyncio
import unittest
import unittest.mock
import weakref
from awebus import EventMixin


class NoWeakrefsTestCase( unittest.TestCase ):
    """
    Perform tests on the EventMixin class where weakrefs are disabled.
    """

    def setUp( self ):
        pass

    def tearDown( self ):
        pass

    def testInvokeOneFunctionOnce( self ):
        bus = EventMixin( event_use_weakref=False )
        handler = unittest.mock.Mock( return_value='Hello, World!', spec=object )
        bus.on( 'fizz', handler )
        results = bus.emit( 'fizz' )
        handler.assert_called_once_with()
        self.assertEqual( results, [ 'Hello, World!' ] )

    def testInvokeDeletedFunctionOnce( self ):
        bus = EventMixin( event_use_weakref=False )
        mock = unittest.mock.Mock( return_value=134, spec=object )

        def handler():
            return mock()

        bus.on( 'buzz', handler )
        del handler
        results = bus.emit( 'buzz' )
        mock.assert_called_once_with()
        self.assertEqual( results, [ 134 ] )



#!/usr/bin/env python

import asyncio
import unittest
import unittest.mock
import weakref
from awebus import EventMixin


class InvokeTestCase( unittest.TestCase ):

    def setUp( self ):
        pass

    def tearDown( self ):
        pass

    def testInvokeOneFunctionOnce( self ):
        bus = EventMixin()
        handler = unittest.mock.Mock( return_value='Hello, World!', spec=object )
        bus.on( 'fizz', handler )
        results = bus.emit( 'fizz' )
        handler.assert_called_with()
        self.assertEqual( [ 'Hello, World!' ], results )

    def testInvokeOneFunctionOnceWithArguments( self ):
        bus = EventMixin()
        handler = unittest.mock.Mock( spec=object )
        bus.on( 'buzz', handler )
        bus.emit( 'buzz', 'a', 'b', 'c' )
        handler.assert_called_with( 'a', 'b', 'c' )

    def testInvokeOneFunctionTwice( self ):
        bus = EventMixin()
        handler = unittest.mock.Mock( spec=object )
        bus.on( 'bleepo', handler )
        bus.on( 'bleepo', handler )
        bus.emit( 'bleepo' )
        self.assertEqual( 2, handler.call_count )

    def testInvokeOneFunctionTwiceWithArguments( self ):
        bus = EventMixin()
        handler = unittest.mock.Mock( spec=object )
        bus.on( 'slammo', handler )
        bus.on( 'slammo', handler )
        bus.emit( 'slammo', 1, door='shut' )
        self.assertEqual( 2, handler.call_count )
        handler.assert_called_with( 1, door='shut' )

    def testInvokeTwoFunctionsOnceTogether( self ):
        bus = EventMixin()
        handler1 = unittest.mock.Mock( return_value='Harry', spec=object )
        handler2 = unittest.mock.Mock( return_value='Sally', spec=object )
        bus.on( 'shout', handler1 )
        bus.on( 'shout', handler2 )
        results = bus.emit( 'shout' )
        handler1.assert_called_with()
        handler2.assert_called_with()
        self.assertEqual( [ 'Harry', 'Sally' ], results )

    def testInvokeTwoFunctionsOnceIndependently( self ):
        bus = EventMixin()
        handler1 = unittest.mock.Mock( spec=object )
        handler2 = unittest.mock.Mock( spec=object )
        bus.on( 'scream', handler1 )
        bus.on( 'whisper', handler2 )
        bus.emit( 'scream' )
        handler1.assert_called_with()
        self.assertEqual( 0, handler2.call_count )
        # handler2.assert_not_called()
        handler1.reset_mock()
        handler2.reset_mock()
        bus.emit( 'whisper' )
        self.assertEqual( 0, handler1.call_count )
        # handler1.assert_not_called()
        handler2.assert_called_with()

    def testInvokeAfterDeletion( self ):
        bus = EventMixin()
        handler = unittest.mock.Mock( spec=object )
        bus.on( 'snarf', handler )
        del handler
        results = bus.emit( 'snarf' )
        self.assertEqual( [ ], results )

    def testInvokeLambda( self ):
        bus = EventMixin()
        handler_mock = unittest.mock.Mock( return_value='did-the-thing', spec=object )
        handler = lambda: handler_mock()
        bus.on( 'do-the-thing', handler )
        results = bus.emit( 'do-the-thing' )
        handler_mock.assert_called_with()
        self.assertEqual( [ 'did-the-thing' ], results )

    def testInvokeSyncHandlersAsynchronously( self ):
        bus = EventMixin()
        handler1 = unittest.mock.Mock( return_value=7, spec=object )
        handler2 = unittest.mock.Mock( return_value=6, spec=object )
        bus.on( 'all-together-now', handler1, handler2 )
        loop = asyncio.get_event_loop()
        results = loop.run_until_complete( bus.emitAsync( 'all-together-now' ) )
        handler1.assert_called_with()
        handler2.assert_called_with()
        self.assertEqual( [ 7, 6 ], results )

    def testInvokeMixedHandlersAsynchronously( self ):
        bus = EventMixin()
        handler1 = unittest.mock.Mock( return_value=7, spec=object )
        handler2 = unittest.mock.Mock( return_value=6, spec=object )

        @asyncio.coroutine
        def handler2async():
            return handler2()

        bus.on( 'all-together-now', handler1, handler2async )

        loop = asyncio.get_event_loop()
        results = loop.run_until_complete( bus.emitAsync( 'all-together-now' ) )
        handler1.assert_called_with()
        handler2.assert_called_with()
        self.assertEqual( [ 7, 6 ], results )

    def testInvokeAsyncHandlersAsynchronously( self ):
        bus = EventMixin()
        handler1 = unittest.mock.Mock( return_value=7, spec=object )
        handler2 = unittest.mock.Mock( return_value=6, spec=object )

        @asyncio.coroutine
        def handler1async():
            return handler1()

        @asyncio.coroutine
        def handler2async():
            return handler2()

        bus.on( 'all-together-now', handler1async, handler2async )
        loop = asyncio.get_event_loop()
        results = loop.run_until_complete( bus.emitAsync( 'all-together-now' ) )
        handler1.assert_called_with()
        handler2.assert_called_with()
        self.assertEqual( [ 7, 6 ], results )

    def testInvokeMixedHandlersSynchronously( self ):
        bus = EventMixin()
        handler1 = unittest.mock.Mock( return_value=7, spec=object )
        mock2 = unittest.mock.Mock( return_value=6, spec=object )

        @asyncio.coroutine
        def handler2():
            return mock2()

        bus.on( 'all-together-now', handler1, handler2 )
        results = bus.emit( 'all-together-now' )
        handler1.assert_called_with()
        mock2.assert_called_with()
        self.assertEqual( [ 7, 6 ], results )

    def testInvokeAsyncHandlersSynchronously( self ):
        bus = EventMixin()
        handler1 = unittest.mock.Mock( return_value=7, spec=object )
        handler2 = unittest.mock.Mock( return_value=6, spec=object )

        @asyncio.coroutine
        def handler1async():
            return handler1()

        @asyncio.coroutine
        def handler2async():
            return handler2()

        bus.on( 'all-together-now', handler1async, handler2async )
        results = bus.emit( 'all-together-now' )
        handler1.assert_called_once_with()
        handler2.assert_called_once_with()
        self.assertEqual( [ 7, 6 ], results )

    def testClassMethodHandlers( self ):
        class C( object ):
            def __init__( self, mock ):
                self.mock = mock

            def onEvent( self ):
                self.mock()

        mock = unittest.mock.Mock( spec=object )
        bus = EventMixin()
        c = C( mock )
        bus.on( 'evt', c.onEvent )
        bus.emit( 'evt' )
        mock.assert_called_once_with()

    def testRemoveClassMethodHandlers( self ):
        class C( object ):
            def __init__( self, mock ):
                self.mock = mock

            def onEvent( self ):
                self.mock()

        mock = unittest.mock.Mock( spec=object )
        bus = EventMixin()
        c = C( mock )
        bus.on( 'evt', c.onEvent )
        bus.off( 'evt', c.onEvent )
        bus.emit( 'evt' )
        self.assertEqual( 0, mock.call_count )



#!/usr/bin/env python

import unittest
import unittest.mock
import weakref
from awebus import EventMixin


class RegistrationTestCase( unittest.TestCase ):

    def setUp( self ):
        pass

    def tearDown( self ):
        pass

    def testRegisterEventHandler( self ):
        bus = EventMixin()
        handler = unittest.mock.Mock( spec = object )
        bus.on( 'foo', handler )
        self.assertIn( 'foo', bus._EventMixin__handlers )
        self.assertIn( weakref.ref( handler ), bus._EventMixin__handlers[ 'foo' ] )
        self.assertEqual( 1, len( bus._EventMixin__handlers[ 'foo' ] ) )

    def testUnregisterEventHandler( self ):
        bus = EventMixin()
        handler = unittest.mock.Mock( spec = object )
        bus.on( 'foo', handler )
        bus.off( 'foo', handler )
        self.assertNotIn( 'foo', bus._EventMixin__handlers )

    # TODO: Add test method to check for unicode event names?



import weakref


class WeakMethod( object ):
    """
    Weak method wrapper to prevent bound methods from disappearing.

    Inspired by: http://code.activestate.com/recipes/81253/
    Changes:
    - Uses Python 3 Syntax
    - Implements __eq__ to allow for equality testing similar to
      the way that you can compare weakref.ref() instances.
    """

    def __init__( self, f ):
        self.f = f.__func__
        self.c = weakref.ref( f.__self__ )

    def __call__( self, *args ):
        c = self.c()
        if c is None:
            return None
        return self.f( c, *args )

    def __eq__( self, other ):
        return other.f == self.f and other.c == self.c

    @property
    def isUnreffed( self ):
        return self.c() is None



#!/usr/bin/env python

import asyncio
import sys
import types
import weakref

from awebus._weak_method import WeakMethod

if sys.version_info < (3, 5):
    def _is_future( o ):
        """
        Return True if the given object is a future(-compatible) object
        """
        return isinstance( o, asyncio.Future )
else:
    def _is_future( o ):
        """
        Return True if the given object is a future(-compatible) object
        """
        return (asyncio.isfuture( o ) or isinstance( o, types.CoroutineType ))


def _is_coro( o ):
    """
    Return True if the given object is a coroutine(-compatible) object
    """
    return asyncio.iscoroutine( o ) or asyncio.iscoroutinefunction( o )


def is_bound( m ):
    """
    Determine if the given object is a bound method
    Source: https://stackoverflow.com/a/18955425
    """
    return hasattr( m, '__self__' )


class EventMixin( object ):
    events = { }

    def __init__( self, *args, **kwargs ):
        self.__handlers = { }
        self.__use_weakref = kwargs.get( 'event_use_weakref', True )
        self.__init_handlers()

    def __init_handlers( self ):
        """
        Register any event handlers that are defined in the __handlers dictionary
        """
        for event, handler_name_or_list in self.events.items():
            if isinstance( handler_name_or_list, str ):
                handler_name_or_list = [ handler_name_or_list ]
            for handler_name in handler_name_or_list:
                handler = getattr( self, handler_name )
                self.on( event, handler )

    def _get_weak( self, o ):
        """
        Retrieve a weak reference to the given object.
        If the object is a bound method, a custom `WeakMethod` wrapper is used
        to keep the object (weakly) alive.

        Params:
            o: The object to retrieve a weakref for.

        Returnsz:
            A weak-reference compatible object.
        """
        if is_bound( o ):
            return WeakMethod( o )
        else:
            return weakref.ref( o )

    def on( self, event, *handlers ):
        """
        Register a handler for an event.

        Args:
            event (str): The name of the event to register handlers for
            handlers (Callable[]): The event handler(s) to invoke when the event is raised

        Returns:
            EventMixin: self
        """
        if event not in self.__handlers:
            self.__handlers[ event ] = [ ]
        for handler in handlers:
            if self.__use_weakref:
                handler = self._get_weak( handler )
            self.__handlers[ event ].append( handler )
        return self

    def off( self, event, *handlers ):
        """
        Unregister a handler for an event.

        Args:
            event (str): The name of the event to register handlers for
            handler (Callable[]): The event handler(s) to invoke when the event is raised

        Returns:
            EventMixin: self
        """
        if event not in self.__handlers:
            return
        for handler in handlers:
            if self.__use_weakref:
                handler = self._get_weak( handler )
            self.__handlers[ event ].remove( handler )
        if len( self.__handlers[ event ] ) == 0:
            del self.__handlers[ event ]
        return self

    def _remove_event_if_empty( self, event ):
        """
        Check the given event and if it has no handlers then remove it from
        the handlers dictionary.

        Args:
            event (str): The name of the event to check

        Returns:
            EventMixin: self
        """
        if event in self.__handlers and len( self.__handlers[ event ] ) == 0:
            del self.__handlers[ event ]
        return self

    def _get_and_clean_event_handlers( self, event ):
        """
        Retrieve concrete implementations of all event handlers, cleaning
        up any unreffed handlers in the process.

        Args:
            event (str): The name of the event to retrieve handlers for.

        Returns:
            Callable[]: A list of concrete event handlers
        """
        if self.__use_weakref:
            handlers = [ ]
            unreffed = [ ]

            # Get handlers from weak references or weak methods
            for weakHandler in self.__handlers.get( event, [ ] ):
                if isinstance( weakHandler, weakref.ref ):
                    handler = weakHandler()
                    if not handler:
                        unreffed.append( weakHandler )
                    else:
                        handlers.append( handler )
                elif isinstance( weakHandler, WeakMethod ):
                    if weakHandler.isUnreffed:
                        unreffed.append( weakHandler )
                    else:
                        handlers.append( weakHandler )

            # Remove any unreffed handlers from the event chain
            if unreffed:
                for handler in unreffed:
                    self.__handlers[ event ].remove( handler )
                self._remove_event_if_empty( event )

            return handlers

        else:
            return self.__handlers.get( event, [ ] )

    def _handler_as_coro( self, o ):
        """
        Coerce any synchronous handlers into an asynchronous coroutine.

        Args:
            o (object): The object to coerce

        Returns:
            object: the original object, or an asynchronous handler if the
            object was not a coroutine.
        """
        if _is_coro( o ):
            return o
        else:
            return asyncio.coroutine( o )

    def emit( self, event, *args, **kwargs ):
        """
        Emit a synchronous event.

        Args:
            event (str): The name of the event to invoke
            args (tuple): Arguments to pass to the event handler(s)
            kwargs (dict): Keyword arguments to pass to the event handler(s)

        Returns:
            list: A list containing the return values of each of the
            executed event handlers.
        """
        results = [ ]
        handlers = self._get_and_clean_event_handlers( event )
        for handler in handlers:
            if _is_coro( handler ):
                loop = asyncio.get_event_loop()
                result = loop.run_until_complete( handler( *args, **kwargs ) )
                results.append( result )
            else:
                result = handler( *args, **kwargs )
                results.append( result )
        return results

    @asyncio.coroutine
    def emitAsync( self, event, *args, **kwargs ):
        """
        Emit an asynchronous event.

        Args:
            event (str): The name of the event to invoke
            args (tuple): Arguments to pass to the event handler(s)
            kwargs (dict): Keyword arguments to pass to the event handler(s)

        Returns:
            awaitable: An awaitable which will resolve to a list containing
            the return values of each of the executed event handlers.
        """
        awaitables = [ ]
        handlers = self._get_and_clean_event_handlers( event )
        for handler in handlers:
            if _is_coro( handler ):
                awaitables.append( handler( *args, **kwargs ) )
            else:
                asyncHandler = asyncio.coroutine( handler )
                awaitables.append( asyncHandler( *args, **kwargs ) )
        result = yield from asyncio.gather( *awaitables )
        return result



#!/usr/bin/env python

import asyncio
import unittest
import unittest.mock
from awebus import EventMixin


class ClassHandlersTestCase( unittest.TestCase ):
    """
    Test the __handlers class-level variable for registration of bus handlers.
    """

    def setUp( self ):
        pass

    def tearDown( self ):
        pass

    def testRegisterAndInvokeOneHandler( self ):
        class C( EventMixin ):
            events = { 'foo': 'onFoo' }

            def __init__( self, mock ):
                EventMixin.__init__( self )
                self.mock = mock

            def onFoo( self ):
                return self.mock()

        mock = unittest.mock.Mock( spec=object )
        c = C( mock )
        c.emit( 'foo' )
        mock.assert_called_once_with()

    def testRegisterAndInvokeMultipleHandlers( self ):
        class C( EventMixin ):
            events = { 'foo': [ 'onFooFirst', 'onFooSecond' ] }

            def __init__( self, mock1, mock2 ):
                EventMixin.__init__( self )
                self.mock1 = mock1
                self.mock2 = mock2

            def onFooFirst( self ):
                return self.mock1()

            def onFooSecond( self ):
                return self.mock2()

        mock1 = unittest.mock.Mock( spec=object )
        mock2 = unittest.mock.Mock( spec=object )
        c = C( mock1, mock2 )
        c.emit( 'foo' )
        mock1.assert_called_once_with()
        mock2.assert_called_once_with()



import os
from distutils.core import setup


def read( file ):
    with open( os.path.join( os.path.dirname( __file__ ), file ) ) as f:
        return f.read()


setup(
    name='awebus',
    version='1.0a6',
    packages=[ 'awebus', ],
    license='MIT',
    description="Asynchronous Weak Event Bus",
    author="Daniel 'Vector' Kerr",
    author_email="admin@vector.id.au",
    url="https://gitlab.com/vectoridau/awebus",
    long_description=read( 'README.md' ),
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
    ],
)
